import unittest

from pyquil import Program, gates

import openqaoa
import openqaoa.qaoa_components.ansatz_constructor.gates as oq_gate_mod
from openqaoa_pyquil.backends.gates_pyquil import PyquilGateApplicator
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle


class TestPyquilGateApplicator(unittest.TestCase):
    def setUp(self):
        available_gates = []
        available_pyquil_gates_name = []
        for each_gate_name, each_gate_function in gates.QUANTUM_GATES.items():
            available_gates.append(each_gate_function)
            available_pyquil_gates_name.append(each_gate_name.lower())
        self.available_pyquil_gates = dict(
            zip(available_pyquil_gates_name, available_gates)
        )

        self.oq_available_gates = (
            tuple(
                [each_gate for each_gate in oq_gate_mod.OneQubitGate.__subclasses__()]
            )
            + tuple(
                [
                    each_gate
                    for each_gate in oq_gate_mod.OneQubitRotationGate.__subclasses__()
                ]
            )
            + tuple(
                [each_gate for each_gate in oq_gate_mod.TwoQubitGate.__subclasses__()]
            )
            + tuple(
                [
                    each_gate
                    for each_gate in oq_gate_mod.TwoQubitRotationGate.__subclasses__()
                ]
            )
        )

        # OneQubitGate, OneQubitRotationGate, TwoQubitGate, TwoQubitRotationGate are not acceptable inputs into the function even though they are of the Gate class
        # RXX, RYY, RZZ, RXY, RZX and RYZ are not supported by Pyquil library
        self.pyquil_excluded_gates = [
            oq_gate_mod.OneQubitGate,
            oq_gate_mod.OneQubitRotationGate,
            oq_gate_mod.TwoQubitGate,
            oq_gate_mod.TwoQubitRotationGate,
            oq_gate_mod.RXX,
            oq_gate_mod.RYY,
            oq_gate_mod.RZZ,
            oq_gate_mod.RXY,
            oq_gate_mod.RZX,
            oq_gate_mod.RYZ,
        ]

    def test_gate_applicator_mapper(self):
        """
        The mapper to the gate applicator should only contain gates that
        are trivially support by the library.
        """

        for each_gate in PyquilGateApplicator.PYQUIL_OQ_GATE_MAPPER.values():
            self.assertTrue(
                each_gate.__name__.lower() in self.available_pyquil_gates.keys(),
                "{}, {}".format(each_gate.__name__, self.available_pyquil_gates.keys()),
            )

    def test_gate_selector(self):
        """
        This method should return the Pyquil Gate object based on the input OQ
        Gate object.
        """

        gate_applicator = PyquilGateApplicator()

        oq_gate_list = tuple(
            [each_gate for each_gate in oq_gate_mod.Gate.__subclasses__()]
        )

        oq_gate_list_1q = oq_gate_list + tuple(
            [each_gate for each_gate in oq_gate_mod.OneQubitGate.__subclasses__()]
        )

        for each_gate in oq_gate_list_1q:
            if each_gate not in self.pyquil_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(applicator=None, qubit_1=None)
                )
                self.assertEqual(
                    self.available_pyquil_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

        oq_gate_list_1qr = oq_gate_list + tuple(
            [
                each_gate
                for each_gate in oq_gate_mod.OneQubitRotationGate.__subclasses__()
            ]
        )

        for each_gate in oq_gate_list_1qr:
            if each_gate not in self.pyquil_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(applicator=None, qubit_1=None, rotation_object=None)
                )
                self.assertEqual(
                    self.available_pyquil_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

        oq_gate_list_2q = oq_gate_list + tuple(
            [each_gate for each_gate in oq_gate_mod.TwoQubitGate.__subclasses__()]
        )

        for each_gate in oq_gate_list_2q:
            if each_gate not in self.pyquil_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(applicator=None, qubit_1=None, qubit_2=None)
                )
                self.assertEqual(
                    self.available_pyquil_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

        oq_gate_list_2qr = oq_gate_list + tuple(
            [
                each_gate
                for each_gate in oq_gate_mod.TwoQubitRotationGate.__subclasses__()
            ]
        )

        for each_gate in oq_gate_list_2qr:
            if each_gate not in self.pyquil_excluded_gates:
                returned_gate = gate_applicator.gate_selector(
                    each_gate(
                        applicator=None,
                        qubit_1=None,
                        qubit_2=None,
                        rotation_object=None,
                    )
                )
                self.assertEqual(
                    self.available_pyquil_gates[returned_gate.__name__.lower()],
                    returned_gate,
                )

    def test_not_supported_gates(self):
        """
        If an unsupported Gate object is passed into the apply_gate method,
        a KeyError should be raised.
        The unsupported Gate object does not exist on the mapper.
        """

        unsupported_list = [
            oq_gate_mod.RXX,
            oq_gate_mod.RYY,
            oq_gate_mod.RZZ,
            oq_gate_mod.RXY,
            oq_gate_mod.RZX,
            oq_gate_mod.RYZ,
        ]
        qubit_1 = 0
        qubit_2 = 1
        rotation_object = None
        circuit = Program()

        gate_applicator = PyquilGateApplicator()

        for each_gate in unsupported_list:
            with self.assertRaises(KeyError):
                gate_applicator.apply_gate(
                    each_gate, qubit_1, qubit_2, rotation_object, circuit
                )

    def test_wrong_argument(self):
        """
        If a supported Gate object is passed into the apply_gate method with the
        incorrect set of arguments, a TypeError should be raised.
        """

        input_gates = {
            tuple(oq_gate_mod.OneQubitRotationGate.__subclasses__()): {
                "wrong_args": [0, 1, None],
                "n_qubits": 1,
            },
            tuple(oq_gate_mod.TwoQubitGate.__subclasses__()): {
                "wrong_args": [0, 1, None],
                "n_qubits": 2,
            },
            tuple(oq_gate_mod.TwoQubitRotationGate.__subclasses__()): {
                "wrong_args": [0, 1, None],
                "n_qubits": 2,
            },
        }
        circuit = Program()
        gate_applicator = PyquilGateApplicator()

        for each_gate_set, input_arguments in input_gates.items():
            for each_gate in each_gate_set:
                each_gate.n_qubits = input_arguments["n_qubits"]
                if each_gate not in self.pyquil_excluded_gates:
                    with self.assertRaises(TypeError):
                        gate_applicator.apply_gate(
                            each_gate, *input_arguments["wrong_args"], circuit
                        )

    def test_wrong_n_qubits(self):
        """
        If a supported Gate object is passed into the apply_gate method with the
        n_qubits attribute that is not 1 or 2, a ValueError should be raised.
        """

        input_gates = {
            tuple(oq_gate_mod.OneQubitRotationGate.__subclasses__()): {
                "args": [0, None],
                "n_qubits": 3,
            }
        }
        circuit = Program()
        gate_applicator = PyquilGateApplicator()

        for each_gate_set, input_arguments in input_gates.items():
            for each_gate in each_gate_set:
                each_gate.n_qubits = input_arguments["n_qubits"]
                if each_gate not in self.pyquil_excluded_gates:
                    with self.assertRaises(ValueError):
                        gate_applicator.apply_gate(
                            each_gate, *input_arguments["args"], circuit
                        )
                    break

    def test_static_methods_1q(self):
        """
        Checks that the static method, apply_1q_rotation_gate, apply the correct
        gate to the circuit object.
        This method is check directly as there are currently no OQ Gate objects
        that call this method through apply_gate.
        """

        gate_applicator = PyquilGateApplicator()
        circuit = Program()
        output_circuit = gate_applicator.apply_1q_fixed_gate(gates.X, 0, circuit)

        self.assertEqual(
            [
                each_instruction.get_qubits()
                for each_instruction in circuit.instructions
            ],
            [{0}],
        )
        self.assertEqual(
            [
                each_instruction.name.lower()
                for each_instruction in circuit.instructions
            ],
            ["x"],
        )

    def test_static_methods_1qr(self):
        """
        Checks that the static method, apply_1q_rotation_gate, apply the correct
        gate to the circuit object.
        """

        gate_applicator = PyquilGateApplicator()
        input_angle = 1
        rot_obj = RotationAngle(lambda x: x, None, input_angle)

        each_sub_gate = [
            each_gate for each_gate in oq_gate_mod.OneQubitRotationGate.__subclasses__()
        ]

        for each_gate in each_sub_gate:
            circuit = Program()
            output_circuit = gate_applicator.apply_gate(
                each_gate(gate_applicator, 0, rot_obj), 0, rot_obj, circuit
            )

            self.assertEqual(
                [each_instruction.params for each_instruction in circuit.instructions],
                [[input_angle]],
            )
            self.assertEqual(
                [
                    each_instruction.get_qubits()
                    for each_instruction in circuit.instructions
                ],
                [{0}],
            )
            self.assertEqual(
                [
                    each_instruction.name.lower()
                    for each_instruction in circuit.instructions
                ],
                [
                    gate_applicator.PYQUIL_OQ_GATE_MAPPER[
                        each_gate.__name__
                    ].__name__.lower()
                ],
            )

    def test_static_methods_2q(self):
        """
        Checks that the static method, apply_2q_fixed_gate, apply the correct
        gate to the circuit object.
        """

        gate_applicator = PyquilGateApplicator()

        each_sub_gate = [
            each_gate for each_gate in oq_gate_mod.TwoQubitGate.__subclasses__()
        ]

        for each_gate in each_sub_gate:
            if each_gate not in self.pyquil_excluded_gates:
                circuit = Program()
                output_circuit = gate_applicator.apply_gate(
                    each_gate(gate_applicator, 0, 1), 0, 1, circuit
                )

                self.assertEqual(
                    [
                        each_instruction.get_qubits()
                        for each_instruction in circuit.instructions
                    ],
                    [{0, 1}],
                )
                self.assertEqual(
                    [
                        each_instruction.name.lower()
                        for each_instruction in circuit.instructions
                    ],
                    [
                        gate_applicator.PYQUIL_OQ_GATE_MAPPER[
                            each_gate.__name__
                        ].__name__.lower()
                    ],
                )

    def test_static_methods_2qr(self):
        """
        Checks that the static method, apply_2q_rotation_gate, apply the correct
        gate to the circuit object.
        """

        gate_applicator = PyquilGateApplicator()
        input_angle = 1
        rot_obj = RotationAngle(lambda x: x, None, input_angle)

        each_sub_gate = [
            each_gate for each_gate in oq_gate_mod.TwoQubitRotationGate.__subclasses__()
        ]

        for each_gate in each_sub_gate:
            if each_gate not in self.pyquil_excluded_gates:
                circuit = Program()
                output_circuit = gate_applicator.apply_gate(
                    each_gate(gate_applicator, 0, 1, rot_obj), 0, 1, rot_obj, circuit
                )

                self.assertEqual(
                    [
                        each_instruction.params
                        for each_instruction in circuit.instructions
                    ],
                    [[input_angle]],
                )
                self.assertEqual(
                    [
                        each_instruction.get_qubits()
                        for each_instruction in circuit.instructions
                    ],
                    [{0, 1}],
                )
                self.assertEqual(
                    [
                        each_instruction.name.lower()
                        for each_instruction in circuit.instructions
                    ],
                    [
                        gate_applicator.PYQUIL_OQ_GATE_MAPPER[
                            each_gate.__name__
                        ].__name__.lower()
                    ],
                )


if __name__ == "__main__":
    unittest.main()



from typing import Callable

from pyquil import Program, gates
from pyquil.quilatom import QubitPlaceholder

from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
import openqaoa.qaoa_components.ansatz_constructor.gates as gates_core


class PyquilGateApplicator(gates_core.GateApplicator):
    PYQUIL_OQ_GATE_MAPPER = {
        gates_core.X.__name__: gates.X,
        gates_core.RZ.__name__: gates.RZ,
        gates_core.RX.__name__: gates.RX,
        gates_core.RY.__name__: gates.RY,
        gates_core.CX.__name__: gates.CNOT,
        gates_core.CZ.__name__: gates.CZ,
        gates_core.CPHASE.__name__: gates.CPHASE,
        gates_core.RiSWAP.__name__: gates.XY,
    }

    library = "pyquil"

    def create_quantum_circuit(self, n_qubits) -> Program:
        """
        Function which creates and empty circuit specific to the pyquil backend.
        Needed for SPAM twirling but can be used more generally.
        """
        return Program()

    def gate_selector(self, gate: gates_core.Gate) -> Callable:
        selected_pyquil_gate = PyquilGateApplicator.PYQUIL_OQ_GATE_MAPPER[gate.__name__]
        return selected_pyquil_gate

    @staticmethod
    def apply_1q_rotation_gate(
        pyquil_gate, qubit_1: int, rotation_object: RotationAngle, circuit: Program
    ) -> Program:
        circuit += pyquil_gate(rotation_object.rotation_angle, qubit_1)
        return circuit

    @staticmethod
    def apply_2q_rotation_gate(
        pyquil_gate,
        qubit_1: int,
        qubit_2: int,
        rotation_object: RotationAngle,
        circuit: Program,
    ) -> Program:
        circuit += pyquil_gate(rotation_object.rotation_angle, qubit_1, qubit_2)
        return circuit

    @staticmethod
    def apply_1q_fixed_gate(pyquil_gate, qubit_1: int, circuit: Program) -> Program:
        circuit += pyquil_gate(qubit_1)
        return circuit

    @staticmethod
    def apply_2q_fixed_gate(
        pyquil_gate, qubit_1: int, qubit_2: int, circuit: Program
    ) -> Program:
        circuit += pyquil_gate(qubit_1, qubit_2)
        return circuit

    def apply_gate(self, gate: gates_core.Gate, *args):
        selected_pyquil_gate = self.gate_selector(gate)
        if gate.n_qubits == 1:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,rotation_object,circuit)
                return self.apply_1q_rotation_gate(selected_pyquil_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,circuit)
                return self.apply_1q_fixed_gate(selected_pyquil_gate, *args)
        elif gate.n_qubits == 2:
            if hasattr(gate, "rotation_object"):
                # *args must be of the following format -- (qubit_1,qubit_2,rotation_object,circuit)
                return self.apply_2q_rotation_gate(selected_pyquil_gate, *args)
            else:
                # *args must be of the following format -- (qubit_1,qubit_2,circuit)
                return self.apply_2q_fixed_gate(selected_pyquil_gate, *args)
        else:
            raise ValueError("Only 1 and 2-qubit gates are supported.")



import networkx as nw
import unittest

from openqaoa import QAOA
from openqaoa.backends import create_device
from openqaoa.problems import MinimumVertexCover
from openqaoa_pyquil.backends import DevicePyquil
from openqaoa_pyquil.backends import QAOAPyQuilWavefunctionSimulatorBackend


class TestingVanillaQAOA(unittest.TestCase):
    def test_set_backend_properties_check_backend_pyquil_statevector(self):
        """
        Check if the backend returned by set_backend_properties is correct
        Based on the input device. For pyquil statevector simulator.
        Also Checks if defaults from workflows are used in the backend.
        """

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        q = QAOA()
        q.set_device(
            create_device(location="local", name="pyquil.statevector_simulator")
        )
        q.compile(problem=qubo_problem)

        self.assertEqual(type(q.backend), QAOAPyQuilWavefunctionSimulatorBackend)

        self.assertEqual(q.backend.init_hadamard, True)
        self.assertEqual(q.backend.prepend_state, None)
        self.assertEqual(q.backend.append_state, None)
        self.assertEqual(q.backend.cvar_alpha, 1)

        self.assertRaises(AttributeError, lambda: q.backend.n_shots)



from .backends import (
    DevicePyquil,
    QAOAPyQuilQPUBackend,
    QAOAPyQuilWavefunctionSimulatorBackend,
)

device_access = {DevicePyquil: QAOAPyQuilQPUBackend}
device_name_to_obj = {
    "pyquil.statevector_simulator": QAOAPyQuilWavefunctionSimulatorBackend
}
device_location = "qcs"
device_plugin = DevicePyquil
device_args = {
    DevicePyquil: [
        "as_qvm",
        "noisy",
        "compiler_timeout",
        "execution_timeout",
        "client_configuration",
        "endpoint_id",
        "engagement_manager",
    ]
}
backend_args = {
    QAOAPyQuilWavefunctionSimulatorBackend: [],
    QAOAPyQuilQPUBackend: [
        "n_shots",
        "active_reset",
        "rewiring",
        "initial_qubit_mapping",
    ],
}



import unittest
import numpy as np

from pyquil import Program, quilbase
from pyquil.gates import RX as p_RX
from pyquil.gates import RY as p_RY
from pyquil.gates import RZ as p_RZ
from pyquil.gates import CZ as p_CZ
from pyquil.gates import CNOT as p_CX
from pyquil.gates import XY as p_XY
from pyquil.gates import CPHASE as p_CPHASE

from openqaoa.qaoa_components.ansatz_constructor.gates import (
    RY,
    RX,
    RZ,
    CZ,
    CX,
    RXX,
    RYY,
    RZZ,
    RZX,
    CPHASE,
    RiSWAP,
)
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
from openqaoa_pyquil.backends.gates_pyquil import PyquilGateApplicator


class TestingGate(unittest.TestCase):
    def setUp(self):
        self.pyquil_gate_applicator = PyquilGateApplicator()

    def test_pyquil_gates_1q(self):
        # Pyquil Gate Applicator
        gate_applicator = self.pyquil_gate_applicator

        # One Qubit Gate Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)

        empty_program = Program()
        llgate = RY(gate_applicator, 0, rotation_angle_obj)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_RY(np.pi, 0))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)

        empty_program = Program()
        llgate = RX(gate_applicator, 0, rotation_angle_obj)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_RX(np.pi, 0))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)

        empty_program = Program()
        llgate = RZ(gate_applicator, 0, rotation_angle_obj)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_RZ(np.pi, 0))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)

    def test_pyquil_gates_2q(self):
        # Pyquil Gate Applicator
        gate_applicator = self.pyquil_gate_applicator

        # Two Qubit Gate Tests
        empty_program = Program()
        llgate = CZ(gate_applicator, 0, 1)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_CZ(0, 1))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)

        empty_program = Program()
        llgate = CX(gate_applicator, 0, 1)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_CX(0, 1))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)

    def test_pyquil_gates_2q_w_gates(self):
        # Pyquil Gate Applicator
        gate_applicator = self.pyquil_gate_applicator

        # Two Qubit Gate with Angles Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)

        empty_program = Program()
        llgate = CPHASE(gate_applicator, 0, 1, rotation_angle_obj)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_CPHASE(np.pi, 0, 1))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)

        empty_program = Program()
        llgate = RiSWAP(gate_applicator, 0, 1, rotation_angle_obj)
        output_program = llgate.apply_gate(empty_program)

        test_program = Program().inst(p_XY(np.pi, 0, 1))

        output_gate_names = [
            instr.name for instr in output_program if type(instr) == quilbase.Gate
        ]
        test_gate_names = [
            instr.name for instr in test_program if type(instr) == quilbase.Gate
        ]
        self.assertEqual(output_gate_names, test_gate_names)


if __name__ == "__main__":
    unittest.main()



from typing import List
from qcs_api_client.client import QCSClientConfiguration
from pyquil.api._engagement_manager import EngagementManager
from pyquil import get_qc

from openqaoa.backends.devices_core import DeviceBase


class DevicePyquil(DeviceBase):
    """
    Contains the required information and methods needed to access remote
    Rigetti QPUs via Pyquil.

    Attributes
    ----------
    n_qubits: `int`
        The maximum number of qubits available for the selected backend.
        Available upon proper initialisation of the class.
    """

    def __init__(
        self,
        device_name: str,
        as_qvm: bool = None,
        noisy: bool = None,
        compiler_timeout: float = 20.0,
        execution_timeout: float = 20.0,
        client_configuration: QCSClientConfiguration = None,
        endpoint_id: str = None,
        engagement_manager: EngagementManager = None,
    ):
        """
        Parameters
        ----------
        device_name: str
            The name of the desired quantum computer. This should correspond to
            a name returned by :py:func:`list_quantum_computers`. Names ending
            in "-qvm" will return a QVM. Names ending in "-pyqvm" will return a
            :py:class:`PyQVM`. Names ending in "-noisy-qvm" will return a QVM
            with a noise model. Otherwise, we will return a QPU with the given
            name.
        as_qvm: bool
            An optional flag to force construction of a QVM (instead of a QPU).
            If specified and set to ``True``, a QVM-backed quantum computer will
            be returned regardless of the name's suffix.
        noisy: bool
            An optional flag to force inclusion of a noise model. If specified
            and set to ``True``, a quantum computer with a noise model will be
            returned regardless of the name's suffix. The generic QVM noise
            model is simple T1 and T2 noise plus readout error. See
            :py:func:`~pyquil.noise.decoherence_noise_with_asymmetric_ro`. Note,
            we currently do not support noise models based on QCS hardware; a
            value of `True`` will result in an error if the requested QPU is a
            QCS hardware QPU.
        compiler_timeout: float
            Time limit for compilation requests, in seconds.
        execution_timeout: float
            Time limit for execution requests, in seconds.
        client_configuration: QCSClientConfiguration
            Optional client configuration. If none is provided, a default one
            will be loaded.
        endpoint_id: str
            Optional quantum processor endpoint ID, as used in the
            `QCS API Docs`_.
        engagement_manager: EngagementManager
            Optional engagement manager. If none is provided, a default one will
            be created.
        """

        self.device_name = device_name
        self.device_location = "qcs"
        self.as_qvm = as_qvm
        self.noisy = noisy
        self.compiler_timeout = compiler_timeout
        self.execution_timeout = execution_timeout
        self.client_configuration = client_configuration
        self.endpoint_id = endpoint_id
        self.engagement_manager = engagement_manager

        self.quantum_computer = get_qc(
            name=self.device_name,
            as_qvm=self.as_qvm,
            noisy=self.noisy,
            compiler_timeout=self.compiler_timeout,
            execution_timeout=self.execution_timeout,
            client_configuration=self.client_configuration,
            endpoint_id=self.endpoint_id,
            engagement_manager=self.engagement_manager,
        )
        self.n_qubits = len(self.quantum_computer.qubits())

    def check_connection(self) -> bool:
        """This method should allow a user to easily check if the credentials
        provided to access the remote QPU is valid.

        If no device was specified in initialisation of object, just runs
        a test connection without a specific device.
        If device was specified, checks if connection to that device
        can be established.

        TODO :
        Accessing Rigetti's QCS is currently unsupported, so this part
        is empty until that is figured out.
        """

        return True

    def connectivity(self) -> List[List[int]]:
        # returns a networkx graph of qubit topology
        G = self.quantum_computer.qubit_topology()
        connectivity_as_list = list(G.edges())
        return connectivity_as_list



import unittest
import numpy as np
import json
import os
import pytest
import subprocess

from openqaoa.backends.qaoa_backend import (
    get_qaoa_backend,
    DEVICE_NAME_TO_OBJECT_MAPPER,
    DEVICE_ACCESS_OBJECT_MAPPER,
)
from openqaoa.qaoa_components import (
    Hamiltonian,
    create_qaoa_variational_params,
    QAOADescriptor,
)


from openqaoa.backends.wrapper import SPAMTwirlingWrapper

from openqaoa.utilities import X_mixer_hamiltonian, bitstring_energy
from openqaoa.backends.qaoa_device import create_device
from openqaoa.backends.basebackend import QAOABaseBackendShotBased


def get_params():
    cost_hamil = Hamiltonian.classical_hamiltonian(
        terms=[[0, 1], [0], [1]], coeffs=[1, 1, 1], constant=0
    )
    mixer_hamil = X_mixer_hamiltonian(2)

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
    variational_params_std = create_qaoa_variational_params(
        qaoa_descriptor, "standard", "ramp"
    )

    return qaoa_descriptor, variational_params_std


class TestingBaseWrapper(unittest.TestCase):
    """
    These tests check that the methods of the wrapper around the backend are working properly.
    """


class TestingSPAMTwirlingWrapper(unittest.TestCase):
    """
    These tests check methods of the SPAM Twirling wrapper.
    """

    @pytest.mark.qvm
    def test_wrap_pyquil_backend(self):
        """
        Testing if the wrapper is backend-agnostic by checking if it can take
        any of the relevant backend objects as an argument.
        """
        rigetti_args = {
            "as_qvm": True,
            "execution_timeout": 10,
            "compiler_timeout": 100,
        }

        device = create_device(location="qcs", name="7q-noisy-qvm", **rigetti_args)

        backend = get_qaoa_backend(
            qaoa_descriptor=self.qaoa_descriptor,
            device=device,
            n_shots=42,
        )
        try:
            SPAMTwirlingWrapper(
                backend,
                n_batches=self.n_batches,
                calibration_data_location=self.calibration_data_location,
            )
        except Exception as e:
            raise ValueError("The {} backend cannot be wrapped.".format(backend), e)

    def setUp(
        self,
    ):
        self.n_batches = 5
        self.calibration_data_location = (
            "./tests/qpu_calibration_data/spam_twirling_mock.json"
        )
        self.qaoa_descriptor, self.variate_params = get_params()

        rigetti_args = {
            "as_qvm": True,
            "execution_timeout": 10,
            "compiler_timeout": 100,
        }

        self.device = create_device(location="qcs", name="7q-noisy-qvm", **rigetti_args)

        self.backend = get_qaoa_backend(
            qaoa_descriptor=self.qaoa_descriptor,
            device=self.device,
            n_shots=42,
        )

        self.wrapped_obj = SPAMTwirlingWrapper(
            self.backend,
            n_batches=self.n_batches,
            calibration_data_location=self.calibration_data_location,
        )

        self.wrapped_obj_trivial = SPAMTwirlingWrapper(
            self.backend,
            n_batches=1,
            calibration_data_location=self.calibration_data_location,
        )

    def test_setUp(self):
        assert (
            self.wrapped_obj.n_batches == self.n_batches
        ), "The number of batches hasn't been set correctly."
        assert (
            self.wrapped_obj.calibration_data_location == self.calibration_data_location
        ), "The location of the calibration file hasn't been set correctly."

    def test_expectation_value_spam_twirled(self):
        """
        Testing the expectation_value_spam_twirled method of the SPAM Twirling
        wrapper in the following scenarious:
        Given very trivial counts where only the 00 state is present and
        calibration factors 1.0, meaning no errors, the expectation value
        of the energy must be the energy of the 00 bitstring.
        Given the same counts but calibration factors 0.5, the expectation
        value of the energy must be twice the energy of the 00 bitstring
        due to corrections coming from the calibration factors.

        """
        try:
            self.wrapped_obj
            self.wrapped_obj_trivial
        except NameError:
            print("Skipping Qiskit tests as Qiskit is not installed.")
            return None

        counts = {"00": 100, "01": 0, "10": 0, "11": 0}
        hamiltonian = Hamiltonian.classical_hamiltonian(
            terms=[[0, 1], [0], [1]], coeffs=[1, 1, 1], constant=0
        )
        calibration_factors = {(1,): 1.0, (0,): 1.0, (0, 1): 1.0}
        assert self.wrapped_obj.expectation_value_spam_twirled(
            counts, hamiltonian, calibration_factors
        ) == bitstring_energy(
            hamiltonian, "00"
        ), "The function which computes the expectation value when using spam twirling with trivial calibration factors doesn't give the correct energy."

        calibration_factors = {(1,): 0.5, (0,): 0.5, (0, 1): 0.5}
        assert self.wrapped_obj.expectation_value_spam_twirled(
            counts, hamiltonian, calibration_factors
        ) == 2 * bitstring_energy(
            hamiltonian, "00"
        ), "The function which computes the expectation value when using spam twirling with calibration factors = 0.5 doesn't give the correct energy."

        counts = {"00": 25, "01": 25, "10": 25, "11": 25}
        calibration_factors = {(1,): 1.0, (0,): 1.0, (0, 1): 1.0}
        assert (
            self.wrapped_obj.expectation_value_spam_twirled(
                counts, hamiltonian, calibration_factors
            )
            == 0
        ), "The function which computes the expectation value when using spam twirling when equal superposition of states doesn't give the correct energy."

        counts = {"00": 13, "01": 26, "10": 42, "11": 66}
        calibration_factors = {(1,): 0.9, (0,): 0.8, (0, 1): 0.7}
        assert (
            self.wrapped_obj.expectation_value_spam_twirled(
                counts, hamiltonian, calibration_factors
            )
            == -0.759502213584
        ), "The function which computes the expectation value when using spam twirling with random counts and factors doesn't give the correct energy."


if __name__ == "__main__":
    unittest.main()
